@@ -1,5 +1,5 @@
 ; Application independent decoder support.
-; Copyright (C) 2000 Red Hat, Inc.
+; Copyright (C) 2000, 2004, 2009 Red Hat, Inc.
 ; This file is part of CGEN.
 ;
 ; This file provides utilities for building instruction set decoders.
@@ -91,11 +91,11 @@
 		 (lambda (elm)
 		   (case (dtable-entry-type elm)
 		     ((insn)
-		      (string-append " " (obj:name (dtable-entry-value elm))))
+		      (stringsym-append " " (obj:name (dtable-entry-value elm))))
 		     ((table)
-		      (string-append " " (subdtable-name (dtable-entry-value elm))))
+		      (stringsym-append " " (subdtable-name (dtable-entry-value elm))))
 		     ((expr)
-		      (string-append " " (exprtable-name (dtable-entry-value elm))))
+		      (stringsym-append " " (exprtable-name (dtable-entry-value elm))))
 		     (else (error "bad dtable entry type:"
 				  (dtable-entry-type elm)))))
 		 (dtable-guts-entries subtable-guts)))))
@@ -145,7 +145,7 @@
 
 (define (-gen-exprtable-name insn-exprs)
   (string-map (lambda (x)
-		(string-append (obj:name (exprtable-entry-insn x))
+		(string-append (obj:str-name (exprtable-entry-insn x))
 			       "-"
 			       (rtx-strdump (exprtable-entry-expr x))))
 	      insn-exprs)
@@ -212,24 +212,60 @@
 	))
 )
 
+; Compute population counts for each bit.  Return it as a vector indexed by bit
+; number.  Rather than computing raw popularity, attempt to compute
+; "disinguishing value" or inverse-entropy for each bit.  The idea is that the
+; larger the number for any particular bit slot, the more instructions it can
+; be used to distinguish.  Raw mask popularity is not enough -- popular masks
+; may include useless "reserved" fields whose values don't change, and thus are
+; useless in distinguishing.
 
-; Compute population counts for each bit.  Return it as a vector indexed by bit number.
-(define (-mask-bit-population masks mask-lens lsb0?)
+(define (-distinguishing-bit-population masks mask-lens values lsb0?)
   (let* ((max-length (apply max mask-lens))
-	 (population (make-vector max-length 0)))
-    (for-each (lambda (mask len)
+	 (0-population (make-vector max-length 0))
+	 (1-population (make-vector max-length 0))
+	 (num-insns (length masks)))
+    ; Compute the 1- and 0-population vectors
+    (for-each (lambda (mask len value)
 		(logit 5 " population count mask=" (number->hex mask) " len=" len "\n")
 		(for-each (lambda (bitno)
-			    (if (bit-set? mask (if lsb0? bitno (- len bitno 1)))
-				(vector-set! population bitno 
-					     (+ 1 (vector-ref population bitno)))))
+			    (let ((lsb-bitno (if lsb0? bitno (- len bitno 1))))
+			      ; ignore this bit if it's not set in the mask
+			      (if (bit-set? mask lsb-bitno)
+				(let ((chosen-pop-vector (if (bit-set? value lsb-bitno)
+							     1-population 0-population)))
+				  (vector-set! chosen-pop-vector bitno 
+					     (+ 1 (vector-ref chosen-pop-vector bitno)))))))
 			  (-range len)))
-	      masks mask-lens)
-    population)
+	      masks mask-lens values)
+    ; Compute an aggregate "distinguishing value" for each bit.
+    (list->vector
+     (map (lambda (p0 p1)
+	    (logit 4 p0 "/" p1 " ")
+	    ; The most useful bits for decoding are those with counts in both
+	    ; p0 and p1. These are the bits which distinguish one insn from
+	    ; another. Assign these bits a high value (greater than num-insns).
+	    ;
+	    ; The next most useful bits are those with counts in either p0
+	    ; or p1.  These bits represent specializations of other insns.
+	    ; Assign these bits a value between 0 and (num-insns - 1). Note that
+	    ; p0 + p1 is guaranteed to be <= num-insns. The value 0 is assigned
+	    ; to bits for which p0 or p1 is equal to num_insns. These are bits
+	    ; which are always 1 or always 0 in the ISA and are useless for
+	    ; decoding purposes.
+	    ;
+	    ; Bits with no count in either p0 or p1 are useless for decoding
+	    ; and should never be considered. Assigning these bits a value of
+	    ; 0 ensures this.
+	    (cond
+	     ((= (+ p0 p1) 0) 0)
+	     ((= (* p0 p1) 0) (- num-insns (+ p0 p1)))
+	     (else (+ num-insns (sqrt (* p0 p1))))))
+	  (vector->list 0-population) (vector->list 1-population))))
 )
 
-
 ; Return a list (0 ... limit-1)
+
 (define (-range limit)
   (let loop ((i 0)
 	     (indices (list)))
@@ -237,14 +273,16 @@
 )
 
 ; Return a list (base ... base+size-1)
+
 (define (-range2 base size)
   (let loop ((i base)
 	     (indices (list)))
     (if (= i (+ base size)) (reverse indices) (loop (+ i 1) (cons i indices))))
 )
 
+; Return a copy of given vector, with all entries with given indices set
+; to `value'
 
-; Return a copy of given vector, with all entries with given indices set to `value'
 (define (-vector-copy-set-all vector indices value)
   (let ((new-vector (make-vector (vector-length vector))))
     (for-each (lambda (index)
@@ -255,19 +293,26 @@
     new-vector)
 )
 
+; Return a list of indices whose counts in the given vector exceed the given
+; threshold.
+; Sort them in decreasing order of populatority.
 
-; Return a list of indices whose counts in the given vector exceed the given threshold.
 (define (-population-above-threshold population threshold)
-  (find (lambda (index) (if (vector-ref population index) 
-			    (>= (vector-ref population index) threshold)
-			    #f))
-	(-range (vector-length population)))
+  (let* ((unsorted
+	  (find (lambda (index) (if (vector-ref population index) 
+				    (>= (vector-ref population index) threshold)
+				    #f))
+		(-range (vector-length population))))
+	 (sorted
+	  (sort unsorted (lambda (i1 i2) (> (vector-ref population i1)
+					    (vector-ref population i2))))))
+    sorted)
 )
 
+; Return the top few most popular indices in the population vector,
+; ignoring any that are already used (marked by #f).  Don't exceed
+; `size' unless the clustering is just too good to pass up.
 
-; Return the top few most popular indices in the population vector, ignoring any
-; that are already used (marked by negative count).  Don't exceed `size' unless
-; the clustering is just too good to pass up.
 (define (-population-top-few population size)
   (let loop ((old-picks (list))
 	     (remaining-population population)
@@ -275,28 +320,36 @@
 					      (vector->list population)))))
       (let* ((new-picks (-population-above-threshold remaining-population count-threshold)))
 	(logit 4 "-population-top-few"
+	       " desired=" size
 	       " picks=(" old-picks ") pop=(" remaining-population ")"
 	       " threshold=" count-threshold " new-picks=(" new-picks ")\n")
 	(cond 
+	 ; No point picking bits with population count of zero.  This leads to
+	 ; the generation of layers of subtables which resolve nothing.  Generating
+	 ; these tables can slow the build by several orders of magnitude.
+	 ((= 0 count-threshold)
+	  (logit 2 "-population-top-few: count-threshold is zero!\n")
+	  old-picks)
 	 ; No new matches?
 	 ((null? new-picks)
-	  (begin (assert (not (null? old-picks)))
-		 old-picks))
+	  (if (null? old-picks)
+	      (logit 2 "-population-top-few: No bits left to pick from!\n"))
+	  old-picks)
 	 ; Way too many matches?
-	 ((> (+ (length new-picks) (length old-picks)) (+ 2 size))
-	  (list-take (+ 2 size) (append new-picks old-picks)))
+	 ((> (+ (length new-picks) (length old-picks)) (+ size 3))
+	  (list-take (+ 3 size) (append old-picks new-picks))) ; prefer old-picks
 	 ; About right number of matches?
-	 ((> (+ (length new-picks) (length old-picks)) (- 1 size))
+	 ((> (+ (length new-picks) (length old-picks)) (- size 1))
 	  (append old-picks new-picks))
 	 ; Not enough?  Lower the threshold a bit and try to add some more.
 	 (else
 	  (loop (append old-picks new-picks)
 		(-vector-copy-set-all remaining-population new-picks #f)
-		(truncate (* 0.8 count-threshold)))))))
+		; Notice magic clustering decay parameter
+		;  vvvv
+		(* 0.75 count-threshold))))))
 )
 
-
-
 ; Given list of insns, return list of bit numbers of constant bits in opcode
 ; that they all share (or mostly share), up to MAX elements.
 ; ALREADY-USED is a list of bitnums we can't use.
@@ -327,9 +380,10 @@
 ; the optimization is left for later.  Also, see preceding FIXME.
 
 (define (decode-get-best-bits insn-list already-used startbit max decode-bitsize lsb0?)
-  (let* ((raw-population (-mask-bit-population (map insn-base-mask insn-list)
-					       (map insn-base-mask-length insn-list)
-					       lsb0?))
+  (let* ((raw-population (-distinguishing-bit-population (map insn-base-mask insn-list)
+							 (map insn-base-mask-length insn-list)
+							 (map insn-value insn-list)
+							 lsb0?))
 	 ; (undecoded (if lsb0?
 	; 		(-range2 startbit (+ startbit decode-bitsize))
 		;	(-range2 (- startbit decode-bitsize) startbit)))
@@ -345,7 +399,6 @@
     sorted-indices)
 )
 
-
 (define (OLDdecode-get-best-bits insn-list already-used startbit max decode-bitsize lsb0?)
   (let ((masks (map insn-base-mask insn-list))
 	; ??? We assume mask lengths are repeatedly used for insns longer
@@ -380,22 +433,31 @@
   (letrec ((opcode (insn-value insn))
 	   (insn-len (insn-base-mask-length insn))
 	   (decode-len (length bitnums))
-	   (compute (lambda (val insn-len decode-len bl)
+	   (compute (lambda (val insn-len decode-len bl default)
 		      ;(display (list val insn-len decode-len bl)) (newline)
 		      ; Oh My God.  This isn't tail recursive.
 		      (if (null? bl)
 			  0
-			  (+ (if (bit-set? val
-					   (if lsb0?
-					       (car bl)
-					       (- insn-len (car bl) 1)))
+			  (+ (if (or (and (>= (car bl) insn-len) (= default 1))
+				     (and (< (car bl) insn-len)
+					  (bit-set? val
+						    (if lsb0?
+							(car bl)
+							(- insn-len (car bl) 1)))))
 				 (integer-expt 2 (- (length bl) 1))
 				 0)
-			     (compute val insn-len decode-len (cdr bl)))))))
-    (let* ((opcode (compute (insn-value insn) insn-len decode-len bitnums))
-	   (opcode-mask (compute (insn-base-mask insn) insn-len decode-len bitnums))
+			     (compute val insn-len decode-len (cdr bl) default))))))
+    (let* ((opcode (compute (insn-value insn) insn-len decode-len bitnums 0))
+	   (opcode-mask (compute (insn-base-mask insn) insn-len decode-len bitnums 1))
 	   (indices (missing-bit-indices opcode-mask (- (integer-expt 2 decode-len) 1))))
-      (logit 3 "insn =" (obj:name insn) " opcode=" opcode " indices=" indices "\n")
+      (logit 3 "insn =" (obj:name insn)
+	     " insn-value=" (number->hex (insn-value insn))
+	     " insn-base-mask=" (number->hex (insn-base-mask insn))
+	     " insn-len=" insn-len
+	     " decode-len=" decode-len
+	     " opcode=" (number->hex opcode)
+	     " opcode-mask=" (number->hex opcode-mask)
+	     " indices=" indices "\n")
       (map (lambda (index) (+ opcode index)) indices)))
 )
 
@@ -410,7 +472,7 @@
 ; Each "slot" is a list of matching instructions.
 
 (define (-fill-slot! insn-vec insn bitnums lsb0?)
-  ;(display (string-append "fill-slot!: " (obj:name insn) " ")) (display bitnums) (newline)
+  ;(display (string-append "fill-slot!: " (obj:str-name insn) " ")) (display bitnums) (newline)
   (let ((slot-nums (-opcode-slots insn bitnums lsb0?)))
     ;(display (list "Filling slot(s)" slot-nums "...")) (newline)
     (for-each (lambda (slot-num)
@@ -467,7 +529,7 @@
 (define -build-decode-table-entry-args #f)
 
 (define (-build-decode-table-entry insn-vec startbit decode-bitsize index index-list lsb0? invalid-insn)
-  (let ((slot (filter-harmlessly-ambiguous-insns (vector-ref insn-vec index))))
+  (let ((slot (vector-ref insn-vec index)))
     (logit 2 "Processing decode entry "
 	   (number->string index)
 	   " in "
@@ -525,42 +587,59 @@
 
 	; If bitnums is still nil there is an ambiguity.
 	(if (null? bitnums)
-
 	    (begin
-	      ; If all insns are marked as DECODE-SPLIT, don't warn.
-	      (if (not (all-true? (map (lambda (insn)
-					 (obj-has-attr? insn 'DECODE-SPLIT))
-				       slot)))
-		  (message "WARNING: Decoder ambiguity detected: "
-			   (string-drop1 ; drop leading comma
-			    (string-map (lambda (insn)
-					  (string-append ", " (obj:name insn)))
-					slot))
-			   "\n"))
-	      ; Things aren't entirely hopeless.  See if any ifield-assertion
-	      ; specs are present.
-	      ; FIXME: For now we assume that if they all have an
-	      ; ifield-assertion spec, then there is no ambiguity (it's left
-	      ; to the programmer to get it right).  This can be made more
-	      ; clever later.
-	      ; FIXME: May need to back up startbit if we've tried to read
-	      ; more of the instruction.
-	      (let ((assertions (map insn-ifield-assertion slot)))
-		(if (not (all-true? assertions))
-		    (begin
-		      ; Save arguments for debugging purposes.
-		      (set! -build-decode-table-entry-args
-			    (list insn-vec startbit decode-bitsize index index-list lsb0? invalid-insn))
-		      (error "Unable to resolve ambiguity (maybe need some ifield-assertion specs?)")))
-		; FIXME: Punt on even simple cleverness for now.
-		(let ((exprtable-entries
-		       (exprtable-sort (map exprtable-entry-make
-					    slot
-					    assertions))))
-		  (dtable-entry-make index 'expr
-				     (exprtable-make
-				      (-gen-exprtable-name exprtable-entries)
-				      exprtable-entries)))))
+	      ; Try filtering out insns which are more general cases of
+	      ; other insns in the slot.  The filtered insns will appear
+	      ; in other slots as appropriate.
+	      (set! slot (filter-non-specialized-ambiguous-insns slot))
+
+	      (if (= 1 (length slot))
+		  ; Only 1 insn left in the slot, so take it.
+		  (dtable-entry-make index 'insn (car slot))
+		  ; There is still more than one insn in 'slot',
+		  ; so there is still an ambiguity.
+		  (begin
+		    ; If all insns are marked as DECODE-SPLIT, don't warn.
+		    (if (not (all-true? (map (lambda (insn)
+					       (obj-has-attr? insn 'DECODE-SPLIT))
+					     slot)))
+			(message "WARNING: Decoder ambiguity detected: "
+				 (string-drop1 ; drop leading comma
+				  (string-map (lambda (insn)
+						(string-append ", " (obj:str-name insn)))
+					      slot))
+				 "\n"))
+			; Things aren't entirely hopeless.  We've warned about
+		        ; the ambiguity.  Now, if there are any identical insns,
+		        ; filter them out.  If only one remains, then use it.
+		    (set! slot (filter-identical-ambiguous-insns slot))
+		    (if (= 1 (length slot))
+			; Only 1 insn left in the slot, so take it.
+			(dtable-entry-make index 'insn (car slot))
+		        ; Otherwise, see if any ifield-assertion
+			; specs are present.
+			; FIXME: For now we assume that if they all have an
+			; ifield-assertion spec, then there is no ambiguity (it's left
+			; to the programmer to get it right).  This can be made more
+			; clever later.
+			; FIXME: May need to back up startbit if we've tried to read
+			; more of the instruction.
+			(let ((assertions (map insn-ifield-assertion slot)))
+			  (if (not (all-true? assertions))
+			      (begin
+				; Save arguments for debugging purposes.
+				(set! -build-decode-table-entry-args
+				      (list insn-vec startbit decode-bitsize index index-list lsb0? invalid-insn))
+				(error "Unable to resolve ambiguity (maybe need some ifield-assertion specs?)")))
+				; FIXME: Punt on even simple cleverness for now.
+			  (let ((exprtable-entries
+				 (exprtable-sort (map exprtable-entry-make
+						      slot
+						      assertions))))
+			    (dtable-entry-make index 'expr
+					       (exprtable-make
+						(-gen-exprtable-name exprtable-entries)
+						exprtable-entries))))))))
 
 	    ; There is no ambiguity so generate the subtable.
 	    ; Need to build `subtable' separately because we
