@@ -1,5 +1,5 @@
 ; Opcode table support.
-; Copyright (C) 2000 Red Hat, Inc.
+; Copyright (C) 2000, 2005, 2009 Red Hat, Inc.
 ; This file is part of CGEN.
 
 ; Append code here to be run before insn parsing/etc.
@@ -162,6 +162,11 @@
 						  (substring syn 2 n)))
 						"), ")))))
 		 (let ((n (id-len (string-drop1 syn))))
+		   (if (= n 0)
+		       (parse-error context "empty or invalid operand name" syntax))
+		   (let ((operand (string->symbol (substring syn 1 (1+ n)))))
+		     (if (not (current-op-lookup operand))
+			 (parse-error context "undefined operand " operand syntax)))
 		   (loop (string-drop (1+ n) syn)
 			 (string-append result op-macro " ("
 					(string-upcase
@@ -197,7 +202,7 @@
   (gen-obj-sanitize
    (ifmt-eg-insn ifmt) ; sanitize based on the example insn
    (string-list
-    "static const CGEN_IFMT " (gen-sym ifmt) " = {\n"
+    "static const CGEN_IFMT " (gen-sym ifmt) " ATTRIBUTE_UNUSED = {\n"
     "  "
     (number->string (ifmt-mask-length ifmt)) ", "
     (number->string (ifmt-length ifmt)) ", "
@@ -214,7 +219,7 @@
 (define (-gen-ifmt-table)
   (string-write
    "/* Instruction formats.  */\n\n"
-   "#define F(f) & @arch@_cgen_ifld_table[CONCAT2 (@ARCH@_,f)]\n\n"
+   (gen-define-with-symcat "F(f) & @arch@_cgen_ifld_table[@ARCH@_" "f]")
    (string-list-map -gen-ifmt-table-1 (current-ifmt-list))
    "#undef F\n\n"
    )
@@ -246,13 +251,13 @@
 
 (define (insn-handlers insn)
   (string-append
-   (number->string (lookup-index 'insn-normal opc-parse-handlers 0))
+   (number->string (assq-lookup-index 'insn-normal opc-parse-handlers 0))
    ", "
-   (number->string (lookup-index 'insn-normal opc-insert-handlers 0))
+   (number->string (assq-lookup-index 'insn-normal opc-insert-handlers 0))
    ", "
-   (number->string (lookup-index 'insn-normal opc-extract-handlers 0))
+   (number->string (assq-lookup-index 'insn-normal opc-extract-handlers 0))
    ", "
-   (number->string (lookup-index 'insn-normal opc-print-handlers 0))
+   (number->string (assq-lookup-index 'insn-normal opc-print-handlers 0))
    )
 )
 
@@ -292,17 +297,18 @@
 
 (define (-gen-insn-enum)
   (logit 2 "Generating instruction enum ...\n")
-  (string-list
-   (gen-enum-decl 'cgen_insn_type "@arch@ instruction types"
-		  "@ARCH@_INSN_"
-		  (cons '(invalid)
-			(append (gen-obj-list-enums (non-multi-insns (current-insn-list)))
-				'((max)))))
-   "/* Index of `invalid' insn place holder.  */\n"
-   "#define CGEN_INSN_INVALID @ARCH@_INSN_INVALID\n\n"
-   "/* Total number of insns in table.  */\n"
-   "#define MAX_INSNS ((int) @ARCH@_INSN_MAX)\n\n"
+  (let ((insns (gen-obj-list-enums (non-multi-insns (current-insn-list)))))
+    (string-list
+     (gen-enum-decl 'cgen_insn_type "@arch@ instruction types"
+		    "@ARCH@_INSN_"
+		    (cons '(invalid) insns))
+     "/* Index of `invalid' insn place holder.  */\n"
+     "#define CGEN_INSN_INVALID @ARCH@_INSN_INVALID\n\n"
+     "/* Total number of insns in table.  */\n"
+     "#define MAX_INSNS ((int) @ARCH@_INSN_"
+     (string-upcase (gen-c-symbol (caar (list-take -1 insns)))) " + 1)\n\n"
    )
+  )
 )
 
 ; Return a reference to the format table entry of INSN.
@@ -350,10 +356,10 @@
   (let* ((all-attrs (current-insn-attr-list))
 	 (num-non-bools (attr-count-non-bools all-attrs)))
     (string-write
+     (gen-define-with-symcat "A(a) (1 << CGEN_INSN_" "a)")
+     (gen-define-with-symcat "OPERAND(op) @ARCH@_OPERAND_" "op")
      "\
-#define A(a) (1 << CONCAT2 (CGEN_INSN_,a))
 #define MNEM CGEN_SYNTAX_MNEMONIC /* syntax value for mnemonic */
-#define OPERAND(op) CONCAT2 (@ARCH@_OPERAND_,op)
 #define OP(field) CGEN_SYNTAX_MAKE_FIELD (OPERAND (field))
 
 /* The instruction table.  */
@@ -376,8 +382,8 @@ static const CGEN_OPCODE @arch@_cgen_insn_opcode_table[MAX_INSNS] =
 };
 
 #undef A
-#undef MNEM
 #undef OPERAND
+#undef MNEM
 #undef OP
 
 "
@@ -403,7 +409,7 @@ static const CGEN_OPCODE @arch@_cgen_insn_opcode_table[MAX_INSNS] =
 
 static int
 asm_hash_insn_p (insn)
-     const CGEN_INSN *insn;
+     const CGEN_INSN *insn ATTRIBUTE_UNUSED;
 {
   return CGEN_ASM_HASH_P (insn);
 }
@@ -453,8 +459,8 @@ asm_hash_insn (mnem)
 
 static unsigned int
 dis_hash_insn (buf, value)
-     const char * buf;
-     CGEN_INSN_INT value;
+     const char * buf ATTRIBUTE_UNUSED;
+     CGEN_INSN_INT value ATTRIBUTE_UNUSED;
 {
   return CGEN_DIS_HASH (buf, value);
 }
@@ -470,10 +476,10 @@ dis_hash_insn (buf, value)
 /* The hash functions are recorded here to help keep assembler code out of
    the disassembler and vice versa.  */
 
-static int asm_hash_insn_p PARAMS ((const CGEN_INSN *));
-static unsigned int asm_hash_insn PARAMS ((const char *));
-static int dis_hash_insn_p PARAMS ((const CGEN_INSN *));
-static unsigned int dis_hash_insn PARAMS ((const char *, CGEN_INSN_INT));
+static int asm_hash_insn_p        (const CGEN_INSN *);
+static unsigned int asm_hash_insn (const char *);
+static int dis_hash_insn_p        (const CGEN_INSN *);
+static unsigned int dis_hash_insn (const char *, CGEN_INSN_INT);
 \n"
    )
 )
@@ -498,7 +504,7 @@ static unsigned int dis_hash_insn PARAMS ((const char *, CGEN_INSN_INT));
     "  {\n"
     "    "
     "-1, " ; macro-insns are not currently enumerated, no current need to
-    "\"" (obj:name minsn) "\", "
+    "\"" (obj:str-name minsn) "\", "
     "\"" (minsn-mnemonic minsn) "\",\n"
     "    " (gen-syntax-entry "MNEM" "OP" (minsn-syntax minsn)) ",\n"
     "    (PTR) & macro_" (gen-sym minsn) "_expansions[0],\n"
@@ -519,7 +525,7 @@ static unsigned int dis_hash_insn PARAMS ((const char *, CGEN_INSN_INT));
     "  {\n"
     "    "
     "-1, " ; macro-insns are not currently enumerated, no current need to
-    "\"" (obj:name minsn) "\", "
+    "\"" (obj:str-name minsn) "\", "
     "\"" (minsn-mnemonic minsn) "\",\n"
     "    " (gen-syntax-entry "MNEM" "OP" (minsn-syntax minsn)) ",\n"
     "    (PTR) & macro_" (gen-sym minsn) "_expansions[0],\n"
@@ -549,7 +555,7 @@ static unsigned int dis_hash_insn PARAMS ((const char *, CGEN_INSN_INT));
 	 (num-non-bools (attr-count-non-bools all-attrs)))
     (string-write
      "/* Formats for ALIAS macro-insns.  */\n\n"
-     "#define F(f) & @arch@_cgen_ifld_table[CONCAT2 (@ARCH@_,f)]\n\n"
+     (gen-define-with-symcat "F(f) & @arch@_cgen_ifld_table[@ARCH@_" "f]")
      (lambda ()
        (string-write-map -gen-ifmt-table-1
 			 (map insn-ifmt (find (lambda (minsn)
@@ -568,10 +574,10 @@ static unsigned int dis_hash_insn PARAMS ((const char *, CGEN_INSN_INT));
 					    (minsn-expansions minsn))
 				"  { 0, 0 }\n};\n\n")))
 			 minsn-list))
+     (gen-define-with-symcat "A(a) (1 << CGEN_INSN_" "a)")
+     (gen-define-with-symcat "OPERAND(op) @ARCH@_OPERAND_" "op")
      "\
-#define A(a) (1 << CONCAT2 (CGEN_INSN_,a))
 #define MNEM CGEN_SYNTAX_MNEMONIC /* syntax value for mnemonic */
-#define OPERAND(op) CONCAT2 (@ARCH@_OPERAND_,op)
 #define OP(field) CGEN_SYNTAX_MAKE_FIELD (OPERAND (field))
 
 /* The macro instruction table.  */
@@ -606,8 +612,8 @@ static const CGEN_OPCODE @arch@_cgen_macro_insn_opcode_table[] =
 };
 
 #undef A
-#undef MNEM
 #undef OPERAND
+#undef MNEM
 #undef OP
 \n"
     ))
@@ -621,9 +627,7 @@ static const CGEN_OPCODE @arch@_cgen_macro_insn_opcode_table[] =
 /* Set the recorded length of the insn in the CGEN_FIELDS struct.  */
 
 static void
-set_fields_bitsize (fields, size)
-     CGEN_FIELDS *fields;
-     int size;
+set_fields_bitsize (CGEN_FIELDS *fields, int size)
 {
   CGEN_FIELDS_BITSIZE (fields) = size;
 }
@@ -632,20 +636,24 @@ set_fields_bitsize (fields, size)
    This plugs the opcode entries and macro instructions into the cpu table.  */
 
 void
-@arch@_cgen_init_opcode_table (cd)
-     CGEN_CPU_DESC cd;
+@arch@_cgen_init_opcode_table (CGEN_CPU_DESC cd)
 {
   int i;
   int num_macros = (sizeof (@arch@_cgen_macro_insn_table) /
 		    sizeof (@arch@_cgen_macro_insn_table[0]));
   const CGEN_IBASE *ib = & @arch@_cgen_macro_insn_table[0];
   const CGEN_OPCODE *oc = & @arch@_cgen_macro_insn_opcode_table[0];
-  CGEN_INSN *insns = (CGEN_INSN *) xmalloc (num_macros * sizeof (CGEN_INSN));
-  memset (insns, 0, num_macros * sizeof (CGEN_INSN));
+  CGEN_INSN *insns = xmalloc (num_macros * sizeof (CGEN_INSN));
+
+  /* This test has been added to avoid a warning generated
+     if memset is called with a third argument of value zero.  */
+  if (num_macros >= 1)
+    memset (insns, 0, num_macros * sizeof (CGEN_INSN));
   for (i = 0; i < num_macros; ++i)
     {
       insns[i].base = &ib[i];
       insns[i].opcode = &oc[i];
+      @arch@_cgen_build_insn_regex (& insns[i]);
     }
   cd->macro_insn_table.init_entries = insns;
   cd->macro_insn_table.entry_size = sizeof (CGEN_IBASE);
@@ -654,7 +662,10 @@ void
   oc = & @arch@_cgen_insn_opcode_table[0];
   insns = (CGEN_INSN *) cd->insn_table.init_entries;
   for (i = 0; i < MAX_INSNS; ++i)
-    insns[i].opcode = &oc[i];
+    {
+      insns[i].opcode = &oc[i];
+      @arch@_cgen_build_insn_regex (& insns[i]);
+    }
 
   cd->sizeof_fields = sizeof (CGEN_FIELDS);
   cd->set_fields_bitsize = set_fields_bitsize;
@@ -679,14 +690,14 @@ void
 (define (cgen-opc.h)
   (logit 1 "Generating " (current-arch-name) "-opc.h ...\n")
   (string-write
-   (gen-copyright "Instruction opcode header for @arch@."
+   (gen-c-copyright "Instruction opcode header for @arch@."
 		  CURRENT-COPYRIGHT CURRENT-PACKAGE)
    "\
 #ifndef @ARCH@_OPC_H
 #define @ARCH@_OPC_H
 
 "
-   (lambda () (gen-extra-opc.h srcdir (current-arch-name))) ; from <arch>.opc
+   (lambda () (gen-extra-opc.h (opc-file-path) (current-arch-name)))
    -gen-insn-enum
    -gen-ifield-decls
    -gen-init-macros
@@ -702,7 +713,7 @@ void
 (define (cgen-opc.c)
   (logit 1 "Generating " (current-arch-name) "-opc.c ...\n")
   (string-write
-   (gen-copyright "Instruction opcode table for @arch@."
+   (gen-c-copyright "Instruction opcode table for @arch@."
 		  CURRENT-COPYRIGHT CURRENT-PACKAGE)
    "\
 #include \"sysdep.h\"
@@ -713,7 +724,7 @@ void
 #include \"@prefix@-opc.h\"
 #include \"libiberty.h\"
 \n"
-   (lambda () (gen-extra-opc.c srcdir (current-arch-name))) ; from <arch>.opc
+   (lambda () (gen-extra-opc.c (opc-file-path) (current-arch-name)))
    -gen-hash-decls
    -gen-ifmt-table
    -gen-insn-opcode-table
